/*      Program driver - preconditioners in pARMS  can be called
 *      through PETSc.
 *----------------------------------------------------------------------
 *  In this test program processor number one reads the whole matrix
 *  from file. The matrix is assumed to be in Harwell-Boeing
 *  format. The problem is solved with PETSc. The preconditioners in
 *  pARMS can be used through PETSc.
 **/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "petscao.h"
#include "petscksp.h"
#include "aux.h"

int main(int argc, char *argv[])
{
  int myid,scale,usmy,ierr,len;
  /* declarations related to Harwell-boeing format for reading the HB
     matri. Second part is related to I/O parameters */
  char matrix[BUFLEN], fname[BUFLEN], *substr, guesol[2], title[72],
    key[8], type[3]; 
  int nrhs, nc, n, nnz, tmp0, tmp, tmp2, tmp3, job;

  /* working array for reading matrix */
  double *a,*rhstmp,*b,*scaletmp, wt;
  int *ja, *ia,*jb,*ib,volume,option[5],wgtflag,numflag, istart, iend;
  int npro, *riord, *iwk, i1,i;
  int mnnz, ii, jj;

  /* working array for symmetryzing matrix */
  double *mc;
  int *jc, *ic;
  int nnzrow, pos;

  Vec       sol, rhs, y;
  Mat       A;
  KSP       ksp, *subksp;
  PC        pc, subpc;
  PetscReal res1, res2, value;
  int       j, iters;

  /* initialize MPI environment */
  PetscInitialize(&argc, &argv, (char *)0, (char *)0);
  MPI_Comm_rank(PETSC_COMM_WORLD, &myid);
  MPI_Comm_size(PETSC_COMM_WORLD, &npro);

  tmp0 = 0;
  nrhs = 1;
  /* Read matrix; either using user-defined function (routine uread) or 
     |               SPARSKIT function for reading Harwell-Boeieng matrices
  */
  /* read matrix name from file input */
  strcpy(matrix, "SHERMAN3");
  /* variable "matrix" stores the name of the file in HB format 
     |
     |   Read a Harwell-Boeing matrix. using wreadmtc c-version of
     |      sparsit routine - call wreadmtc a first time to determine sizes
     |      of arryas. read in values on the second call. 
  */
  len = strlen(matrix);
  a = NULL; ja = NULL; ia = NULL; rhstmp = NULL;
  wreadmtc_(&tmp0,&tmp0,&tmp0,matrix,&len,a,ja,ia,rhstmp,&nrhs,
	   guesol,&n,&nc,&nnz,title,key,type,&ierr,100,2,72,8,3); 
  
  ia = malloc((n+1)*sizeof(*ia));
  ja = malloc(nnz*sizeof(*ja));
  a  = malloc(nnz*sizeof(*a));
  
  if (nrhs) {
    tmp = 3;
    tmp2 = n;
    tmp3 = nnz;
    if (guesol[0] == 'G' || guesol[0] == 'g') nrhs++;
    if (guesol[1] == 'X' || guesol[1] == 'x') nrhs++;
    rhstmp = malloc((n*nrhs)*sizeof(*rhstmp));
  }
  if (!nrhs) {
    tmp = 2;
    tmp2 = n;
    tmp3 = nnz;
  }
  
  /* Array sizes determined. Now call wreadmtc again for really
     reading */
  wreadmtc_(&tmp2,&tmp3,&tmp,matrix,&len,a,ja,ia,rhstmp,&nrhs,
	   guesol,&n,&nc,&nnz,title,key,type,&ierr,100,2,72,8,3);
  free(rhstmp);
  if(ierr > 0 && ierr < 4 && !myid) {
    fprintf(stderr, "cannot read matrix\n");
    fprintf(stderr, "ierr = %d\n", ierr);
    exit(1);
  }
  else if (ierr >= 4 && !myid){
    fprintf(stderr, "*** Warning: RHS or solution vectors are not read. Artificial RHS will be used\n");
    nrhs = 0;
  }
  
  if(myid == 0) {
    fprintf(stdout,"READ the matrix %.*s %.*s \n",8,key,3,type);
  }
  

  if (myid == 0){
    fprintf(stdout,"Matrix dimension is %d, Number of nonzeros is %d\n",n,nnz);
  }
  
  
  /* the matrix is symmetrized */
  b  = malloc(nnz*sizeof(*b));
  jb = malloc(nnz*sizeof(*jb));
  ib = malloc((n+1)*sizeof(*ib));
  
  job = 0;
  i = 1;
  csrcsc_(&n, &job, &i, a, ja, ia, b, jb, ib);
  usmy = 1;
  if(usmy == 1) {
    iwk = malloc(n*sizeof(*iwk));
    mc  = malloc((2*nnz)*sizeof(*mc));
    jc  = malloc((2*nnz)*sizeof(*jc));
    ic  = malloc((n+1)*sizeof(*ic));
    
    for(i = 0; i < nnz; i++) {
      b[i] = 0.0;
    }
    job = 1; 
    i1 = 2*nnz;
    
    /* compute C = A + B */
    aplb_(&n, &n, &job, a, ja, ia, b, jb, ib, mc, jc, ic, &i1, iwk,
	 &ierr);
    if (ierr) fprintf(stderr, "Error: in aplb, ierr = %d\n", ierr);
    nnz = ic[n]-1;
    if (!myid)
      fprintf(stderr,
	      "Matrix pattern has been symmetrized; added %d nonzeros.\n",
	      ic[n]-ia[n]);

    ja = realloc(ja, nnz*sizeof(*ja));
    a  = realloc(a,  nnz*sizeof(*a));

    memcpy(ia, ic, (n+1)*sizeof(*ia));
    memcpy(ja, jc, nnz*sizeof(*ia));
    memcpy(a, mc, nnz*sizeof(*a));

    free(b);
    free(jb);
    free(ib);
    free(mc);
    free(jc);
    free(ic);
    free(iwk);
  }

  /* create Mat object */
  mnnz = 0;
  for (i = 1; i <= n; i++) {
    nnzrow = ia[i] - ia[i-1];
    if (mnnz < nnzrow) {
      mnnz = nnzrow;
    }
  }

  /* create a matrix object */
  MatCreateMPIAIJ(PETSC_COMM_WORLD, PETSC_DECIDE, PETSC_DECIDE, n, n, mnnz, PETSC_NULL, mnnz,
		  PETSC_NULL, &A);
  // insert entries into the matrix
  for(i = 1; i <= n; i++) {
    nnzrow = ia[i] - ia[i-1];
    pos = ia[i-1] - ia[0];
    ii = i-1;
    for (j = 0; j < nnzrow; j++) {
      jj = ja[pos+j] - 1;
      MatSetValues(A, 1, &ii, 1, &jj, &a[pos+j], INSERT_VALUES);
    }
  }

  /* free temporary arrays */
  free(a); free(ja); free(ia);

  /* assemble the matrix */
  MatAssemblyBegin(A, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(A, MAT_FINAL_ASSEMBLY);

  /* create an abstract krylov subspace object */
  KSPCreate(PETSC_COMM_WORLD, &ksp);

  /* create solution and right-hand-side vector */
  VecCreateMPI(PETSC_COMM_WORLD, PETSC_DECIDE, n, &rhs);
  VecDuplicate(rhs, &sol);
  VecDuplicate(rhs, &y);

  value = 1.0;
  VecSet(&value, sol);
  MatMult(A, sol, rhs); 

  /* set initial guess */
  value = 0.0;
  VecSet(&value, sol);

  VecAssemblyBegin(sol);
  VecAssemblyEnd(sol);

  /* initial residual norm */
  MatMult(A, sol, y);
  value = -1.0;
  VecAXPY(&value, rhs, y);
  VecNorm(y, NORM_2, &res1);

  /* solve linear systems with the krylov subspace method selected */
  KSPSetOperators(ksp, A, A, SAME_NONZERO_PATTERN);
  KSPSetType(ksp, KSPFGMRES); 
  KSPGMRESSetRestart(ksp, 60);
  KSPSetTolerances(ksp, 1.0e-6, 1.0e-20, 1.0e5, 500);
  KSPSetRhs(ksp, rhs);
  KSPSetSolution(ksp, sol);

  KSPSetFromOptions(ksp);

  KSPSolve(ksp);
  KSPGetIterationNumber(ksp, &iters);
  VecNorm(rhs, NORM_2, &res1);
  KSPGetResidualNorm(ksp, &res2);
  if(myid == 0) {
    fprintf(stderr, "The number of iteration %5s %d\n", "=",iters);
    fprintf(stderr, "The initial residual error  = %8.3e\n", res1);
    fprintf(stderr, "The solution residual error = %8.3e\n", res2);
  }

  /* free memories for map, vec, mat, solver */
  MatDestroy(A);
  VecDestroy(sol);
  VecDestroy(rhs);
  VecDestroy(y);
  KSPDestroy(ksp);

  /* exit MPI environment */
  PetscFinalize();
}
