/*--------------------------------------------------------------------
  parms_MatApply          : perform the multiplication of
                            matrix-vector product. 
  parms_MatCreate         : create a matrix object.
  parms_MatExtend         : extend submatrix by including equations
                            correspond to the immediate neighbouring
                            variables. 
  parms_MatFree           : free the memory for a matrix object.
  parms_MatFreeSubMat     : free the memory for the submatrix object.
  parms_MatGetCommHandler : get the communication handler.
  parms_MatGetDiag        : get the diagonal part of the local matrix. 
  parms_MatGetSubMat      : get the local matrix.
  parms_MatILU            : perform ILU factorization.
  parms_MatMVPY           : perform \f$z = \alpha \times self \times x
                                           + beta \times y\f$. 
  parms_MatSetCommType    : set the communication type.
  parms_MatSetILUType     : set local ILU type.
  parms_MatSetType        : set the matrix type.
  parms_MatSetValues      : insert/add values to the matrix object.
  parms_MatSetup          : set up a matrix object.
  parms_MatVec            : perform the multiplication of
                            matrix-vector product. 
  parms_MatVecOffDiag     : perform the multiplication of the
                            off-diagonal matrix and the external vars. 
  parms_MatView           : dump a matrix object.

  A code fragment for using matrix functions:

  parms_Mat mat;
  parms_Map map;
  parms_Vec vec, y;
  parms_PC  pc;

  ...
  // create vectors
  parms_VecCreate(&vec, map);
  parms_VecCreate(&y,   map);

  // insert values into vector
  parms_VecSetValues(...);

  // create a matrix object.
  parms_MatCreate(&mat, map);

  // insert/add values into the matrix
  parms_MatSetValues(mat, ...);

  // setup the matrix. 
  // this function divides local variables into two categories:
  // interior variables and interface variables. interior variables
  // listed first. when the matrix-vector product is performed, the
  // vector is permuted automatically if the vector is created
  // (parms_VecCreate) before calling parms_MatSetup.
  parms_MatSetup(mat);

  // matrix-vector product
  parms_MatVec(mat, vec, y);

  // create a preconditioner object.
  parms_PCCreate(&pc, mat);

  $Id: parms_mat.c,v 1.1.1.1 2006/11/24 21:15:29 zli Exp $
  ------------------------------------------------------------------*/
#include "parms_mat_impl.h"

static int parms_Mat_count      = 0;

extern int parms_MatCreate_vcsr(parms_Mat self);
extern int parms_MatCreate_dvcsr(parms_Mat self);

o_func parms_mattab[] = {(o_func)parms_MatCreate_vcsr,
		   (o_func)parms_MatCreate_dvcsr}; 

/** 
 * Free the parms_Mat object pointed to by self.
 * 
 * @param self A pointer to a parms_Mat object.
 * 
 * @return 0 on success.
 */
int parms_MatFree(parms_Mat *self) 
{
  int         m, i, nnz;

  PARMS_VALIDMATPTR(*self);
  PARMS_OBJSUBREF(*self);
  if (PARMS_OBJGETREF(*self) == 0) {
    (*self)->o_vptr->dtor((parms_Object *)self);
    parms_MapFree(&(*self)->is);
    if ((*self)->isalloc) {
      m = (*self)->m;
      for (i = 0; i < m; i++) {
	nnz = (*self)->aux_data->nnzrow[i];
	if (nnz) {
	  PARMS_FREE((*self)->aux_data->pj[i]);
	  PARMS_FREE((*self)->aux_data->pa[i]);
	}
      }
      PARMS_FREE((*self)->aux_data->pa);
      PARMS_FREE((*self)->aux_data->pj);
      PARMS_FREE((*self)->aux_data->nnzrow);
      PARMS_FREE((*self)->aux_data);
      PARMS_OBJFREE(*self);
    }
  }
  return 0;
}

/** 
 * Dump parms_Mat object.
 * 
 * @param self A pointer to a parms_Mat object.
 * @param v    A parms_Viewer object.   
 * 
 * @return 0 on success.
 */
int parms_MatView(parms_Mat self, parms_Viewer v)
{
  PARMS_VALIDMATPTR(self);
  return self->o_vptr->view((parms_Object)self, v);
}

/** 
 * Set the communication type.
 *
 * Set the communication style across processors.
 * communication style:
 *  -P2P       point-to-point (data copied to/from auxilliary buffers).
 *  -DERIVED   derived datatype.
 *
 * @param self  A matrix object.
 * @param ctype Communication style:
 *              - P2P     point-to-point (data copied to/from
 *                        auxilliary buffers).
 *              - DERIVED derived datatype.
 *              
 * @return 0 on success.
 */
int parms_MatSetCommType(parms_Mat self, COMMTYPE ctype)
{
  PARMS_VALIDMATPTR(self);
  return self->d_vptr->setcommtype(self, ctype);
}

/** 
 * Set the matrix type.
 * 
 * @param self A matrix object.
 * @param type The matrix type:
 *             - MAT_VCSR for VCSR format.
 *             - MAT_CSR for CSR format.
 *             
 * @return 0 on success.
 */
int parms_MatSetType(parms_Mat self, MATTYPE type)
{
  o_func f;

  PARMS_VALIDMATPTR(self);
  if (self->type != MAT_NULL) {
    if (self->type != type && self->issetup) {
      self->o_vptr->dtor((parms_Object *)&self);
    }
  }
  self->type = type;
  if (self->isserial) {
    f = parms_mattab[2*type];
  }
  else {
    f = parms_mattab[2*type+1];
  }

  f((parms_Object)self);
  return 0;
}


/** 
 * Create a parms_Mat object.
 *
 * Create a parms_Mat object based on data distribution layout map.
 * 
 * @param self  A pointer to the parms_Mat object created. 
 * @param map   A parms_Map object, which describes the data
 *              distribution among processors.  
 * 
 * @return 0 on success.
 */
int parms_MatCreate(parms_Mat *self, parms_Map map)
{
  parms_Mat  new_mat;

  PARMS_VALIDMAPPTR(map);
  PARMS_OBJCREATE(new_mat, (o_dtor)0, (o_view)0); 
  new_mat->is               = map;
  PARMS_OBJADDREF(map);
  new_mat->type             = MAT_NULL;
  new_mat->issetup          = false;
  new_mat->isperm           = map->isperm;
  new_mat->isalloc          = false;

  new_mat->m = parms_MapGetLocalSize(map);
  new_mat->n = new_mat->m;
  new_mat->M = parms_MapGetGlobalSize(map);
  new_mat->N = new_mat->M;
  new_mat->isserial = map->isserial;
  new_mat->handler = PARMS_OBJECT_SET_CLASS(PARMS_MAT_CLASS) +
    parms_Mat_count++; 
  *self = new_mat;
  return 0;
}

/** 
 * Perform \f$y = self \times x\f$.
 * 
 * @param self A parms_Mat object.      
 * @param x    A parms_Vec object.      
 * @param y    Another parms_Vec object.
 * 
 * @return 0 on success.
 */
int parms_MatApply(parms_Mat self, parms_Vec x, parms_Vec y)
{
  PARMS_VALIDMATPTR(self);
  PARMS_VALIDVECPTR(x);
  PARMS_VALIDVECPTR(y);
  return self->d_vptr->apply(self, x, y);
}

/** 
 * Perform \f$y = self \times x\f$.
 * 
 * @param self A parms_Mat object.      
 * @param x    A parms_Vec object.      
 * @param y    Another parms_Vec object.
 * 
 * @return 0 on success.
 */
int parms_MatVec(parms_Mat self, parms_Vec x, parms_Vec y)
{
  return parms_MatApply(self, x, y);
}

/** 
 * Set up parms_Mat object self.
 *
 * This is the most important functoin for the parms_Mat object. This
 * function combines the function bdry and setup in the old version 
 * of pARMS. The function sets up the data structure needed by the
 * distributed matrix-vector multiplication, divides the variables on
 * the local processors into two categories: interior and interface
 * variables.
 * 
 * @param self A parms_Mat object. 
 * 
 * @return 0 on success.
 */
int parms_MatSetup(parms_Mat self)
{
  PARMS_VALIDMATPTR(self);
  if ((!self->issetup) && (self->type == MAT_NULL)) {
    parms_MatSetType(self, MAT_VCSR);
  }
  return self->d_vptr->setup(self);
}

/** 
 * Insert/add values to the parms_Mat object self.
 * 
 * @param self    A parms_Mat object.                                
 * @param m 	The number of rows inserted.                       
 * @param im 	An array of row global indices.                    
 * @param ia 	An array of pointer to the beginning of each row in
 *                array ja.    
 * @param ja      An array of column global indices.     
 * @param values  An array of values.                    
 * @param mode 	Insert value mode:                     
 * 		      - INSERT  insert values to parm_Mat self.
 * 		      - ADD     add values to parm_Mat self.
 * 		  
 * @return 0 on success.
 */
int parms_MatSetValues(parms_Mat self, int m, int *im, int *ia,
		       int *ja, FLOAT *values, INSERTMODE mode)  
{
  parms_Map  is;
  parms_vcsr aux_data;
  BOOL isserial, isalloc;
  int i, j, k, offset, start, end, rindex, cindex;
  int space, numinf, index, lrindex, lcindex, *rja;
  int *perm, *rp, *cp, *cep, lsize, size;
  FLOAT *ra;
  BOOL found;

  PARMS_VALIDMATPTR(self);
  isalloc  = self->isalloc;
  is       = self->is;
  offset    = is->start;
  lsize    = parms_MapGetLocalSize(is);
  isserial = is->isserial;
  if (isalloc) {
    aux_data = self->aux_data;
  }
  else {
    self->isalloc = true;
    PARMS_NEW(aux_data);
    aux_data->n = lsize;
    PARMS_NEWARRAY(aux_data->space,   aux_data->n);
    PARMS_NEWARRAY0(aux_data->nnzrow, aux_data->n);
    PARMS_NEWARRAY(aux_data->pj,      aux_data->n);
    PARMS_NEWARRAY(aux_data->pa,      aux_data->n);
    for (i = 0; i < aux_data->n; i++) {
      aux_data->space[i] = 30;
      PARMS_NEWARRAY(aux_data->pj[i], aux_data->space[i]);
      PARMS_NEWARRAY(aux_data->pa[i], aux_data->space[i]);
    }
    self->aux_data = aux_data;
  }

  if (isserial) {
    for (i = 0; i < m; i++) {
      rindex = im[i] - offset;
      ra     = aux_data->pa[rindex];
      rja    = aux_data->pj[rindex];
      start  = ia[i] - offset;
      end    = ia[i+1] - offset;
      for (j = start; j < end; j++) {
	cindex = ja[j] - offset;
	space = aux_data->space[rindex];
	if (space > aux_data->nnzrow[rindex]) {
	  found = false;
	  for (k = 0; k < aux_data->nnzrow[rindex]; k++) {
	    if (rja[k] == cindex) {
	      found = true;
	      index = k;
	      break;
	    }
	  }
	  if (found) {
	    if (mode == INSERT) {
	      ra[index] = values[j];
	    }
	    else if (mode == ADD) {
	      ra[index] += values[j];
	    }
	  }
	  else { /* insert a new entry */
	    rja[aux_data->nnzrow[rindex]]  = cindex;
            ra[aux_data->nnzrow[rindex]++] = values[j];
	  }
	}
	else { /* reallocate memory for holding new entry */
	  space += 10;
	  PARMS_RESIZE(aux_data->pa[rindex], space);
	  PARMS_RESIZE(aux_data->pj[rindex], space);
	  
	  ra  = aux_data->pa[rindex];
	  rja = aux_data->pj[rindex];
	  
	  aux_data->space[rindex]        = space;
	  rja[aux_data->nnzrow[rindex]]  = cindex;
          ra[aux_data->nnzrow[rindex]++] = values[j];
	}
      }
    }
  }
  else {
    perm    = is->perm;
    numinf  = is->ninf;

    for (i = 0; i < m; i++) {
      rindex = im[i] - offset;
      rp = parms_MapGlobalToLocal(is, rindex);
      /* this row resides the local processor */
      if (rp != NULL && *rp < lsize) {
	// calculate the local row index
	lrindex = *rp;
	ra      = aux_data->pa[lrindex];
	rja     = aux_data->pj[lrindex];

	start = ia[i]-offset;
	end   = ia[i+1]-offset;
	for (j = start; j < end; j++) {
	  cindex = ja[j] - offset;
/*-------------------- the local processor contains this column index */
	  cp = parms_MapGlobalToLocal(is, cindex);
	  if (cp != NULL && *cp < lsize) {
	    /* local column index */
	    lcindex = *cp;
	    space   = aux_data->space[lrindex];
	    if (space > aux_data->nnzrow[lrindex]) {
	      found = false;
	      for (k = 0; k < aux_data->nnzrow[lrindex]; k++) {
		if (rja[k] == lcindex) {
		  found = true;
		  index = k;
		  break;
		}
	      }
	      if (found) {
		if (mode == INSERT) {
		  ra[index] = values[j];
		}
		else if (mode == ADD) {
		  ra[index] += values[j];
		}
	      }
	      else { /* insert the new entry */
		rja[aux_data->nnzrow[lrindex]]  = lcindex;
		ra[aux_data->nnzrow[lrindex]++] = values[j];
	      }
	    }
	    else { /* reallocate memory for holding new entry */
	      space += 10;
	      PARMS_RESIZE(aux_data->pa[lrindex], space);
	      PARMS_RESIZE(aux_data->pj[lrindex], space);
	      ra  = aux_data->pa[lrindex];
	      rja = aux_data->pj[lrindex];

	      aux_data->space[lrindex]        = space;
	      rja[aux_data->nnzrow[lrindex]]  = lcindex;
	      ra[aux_data->nnzrow[lrindex]++] = values[j];
	    }
	  }
	  /*-------------- column index doesn't reside on the local processor */
	  else { 
	    if (perm[lrindex] == -1) { /* nor marked yet */
	      perm[lrindex] = lsize-1-numinf;
	      numinf++;
	    }
	    space = aux_data->space[lrindex];
	    if (space > aux_data->nnzrow[lrindex]) {
	      found = false;
	      for (k = 0; k < aux_data->nnzrow[lrindex]; k++) {
		if (rja[k] == -cindex-1) {
		  found = true;
		  index = k;
		  break;
		}
	      }
/*-------------------- case when entry is found */
	      if (found) {
		if (mode == INSERT) {
		  ra[index] = values[j];
		}
		else if (mode == ADD) {
		  ra[index] += values[j];
		}
	      }
	      else {
		cep = parms_MapGlobalToLocal(is, cindex);
		if (cep == NULL) {
		  size = parms_TableGetSize(is->table);
		  parms_TablePut(is->table, cindex, size);
		}
		rja[aux_data->nnzrow[lrindex]]  = -cindex-1;
		ra[aux_data->nnzrow[lrindex]++] = values[j];
	      }
	    }
	    else {
/*-------------------- reallocate case */
	      space += 10;
	      PARMS_RESIZE(aux_data->pa[lrindex], space);
	      PARMS_RESIZE(aux_data->pj[lrindex], space);
	      ra  = aux_data->pa[lrindex];
	      rja = aux_data->pj[lrindex];
	      aux_data->space[lrindex]        = space;
	      cep = parms_MapGlobalToLocal(is, cindex);
	      if (cep == NULL) {
		size = parms_TableGetSize(is->table);
		parms_TablePut(is->table, cindex, size);
	      }
	      rja[aux_data->nnzrow[lrindex]]  = -cindex-1;
	      ra[aux_data->nnzrow[lrindex]++] = values[j];
	    }
	  }
	}
      }
    }
    is->ninf = numinf;
  }

  return 0;
}

/** 
 * Set local ILU type.
 * 
 * @param self A matrix object. 
 * @param type ILU type:
 *	       -PCILU0  
 *	       -PCILUK  
 *	       -PCILUT  
 *	       -PCARMS  
 * 
 * @return 0 on success.
 */
int parms_MatSetILUType(parms_Mat self, PCILUTYPE type)
{
  PARMS_VALIDMATPTR(self);
  return self->d_vptr->setilu(self, type);
}

/** 
 * Get the diagonal part of the local matrix. 
 * 
 * @param self A parms_Mat object.
 * @param mat  The diagonal part of the local matrix.    
 * 
 * @return 0 on success.
 */
int parms_MatGetDiag(parms_Mat self, void **mat)
{
  PARMS_VALIDMATPTR(self);
  return self->d_vptr->getdiag(self, mat);
}

/** 
 * Perform ILU factorization.
 * 
 * @param self   A matrix object.                                  
 * @param param  The parameters used for ILU factorization.        
 * @param mat    Matrix to be factored. 
 * @param op     The operator created.   
 * 
 * @return 0 on success.
 */
int parms_MatILU(parms_Mat self, parms_FactParam param, void *mat,
		 parms_Operator *op) 
{
  PARMS_VALIDMATPTR(self);
  return self->d_vptr->ilu(self, param, mat, op);
}

/** 
 * Perform \f$z = alpha*self*x + beta*y\f$.
 * 
 * @param self   A matrix object.           
 * @param alpha  A scalar.                  
 * @param x      A vector object.           
 * @param beta   A scalar.                  
 * @param y      A vector object.           
 * @param z      A vector stores the result.
 * 
 * @return 0 on success.
 */
int parms_MatMVPY(parms_Mat self, REAL alpha, parms_Vec x, REAL beta,
		  parms_Vec y, parms_Vec z)
{
  PARMS_VALIDMATPTR(self);
  PARMS_VALIDVECPTR(x);
  PARMS_VALIDVECPTR(y);
  PARMS_VALIDVECPTR(z);

  return self->d_vptr->mvpy(self, alpha, x, beta, y, z);
}

/** 
 * Perform the multiplication of the off-diagonal matrix and the
 * external vars. 
 *
 * The local matrix can be written as follows:
 * 
 *  \f[
 *  \left(
 *  \begin{array}{ccc}
 *    B   &   E & 0\\
 *    F   &   C & M_{ext}
 *  \end{array}
 *  \right)
 *  \f],
 *  where \f$\left(\begin{array}{cc}
 *    B  &   E \\
 *    F  &   C 
 *  \end{array}
 *  \right)\f$ corresponds to the variables on the local PE. This
 *  function performs 
 *  \f[
 *    y[pos..n] = M_{ext} \times x_{ext}
 *  \f]
 *
 * @param self A matrix object.                                      
 * @param x    A vector object.                                      
 * @param y    A vector object.                                      
 * @param pos  The offset of x from the beginning of the local
 *             vector. 
 * 
 * @return 0 on success.
 */
int parms_MatVecOffDiag(parms_Mat self, FLOAT *x, FLOAT *y, int pos)
{
  PARMS_VALIDMATPTR(self);

  return self->d_vptr->mvoffd(self, x, y, pos);
}

/** 
 * Free the memory for the submatrix.
 * 
 * @param self A parms_Mat object.             
 * @param mat  The submatrix to be freed.      
 * 
 * @return 0 on success.
 */
int parms_MatFreeSubMat(parms_Mat self, void *mat)
{
  PARMS_VALIDMATPTR(self);

  return self->d_vptr->matfree(self, mat);
}

/** 
 * Extend submatrix by including equations correspond to the
 * immediate neighbouring variables.
 * 
 * @param self     A matrix object.                                         
 * @param handler  A communication handler.                                 
 * @param start    The beginning location of mat in the local matrix.       
 * @param mat      The submatrix to be extended.        
 * @param n 	 The size of extended matrix returned.
 * @param ext_mat  The extended matrix created.            
 * 
 * @return 0 on success.
 */
int parms_MatExtend(parms_Mat self, parms_Comm handler, int start,
		    void *mat, int *n, void **ext_mat)
{
  PARMS_VALIDMATPTR(self);
  PARMS_VALIDCOMMPTR(handler);

  return self->d_vptr->extend(self, handler, start, mat, n, ext_mat);
}

/** 
 * Get the local matrix. 
 * 
 * @param self A matrix object.                               
 * @param mat  The submatrix returned in a specific format.   
 * 
 * @return 0 on success.
 */
int parms_MatGetSubMat(parms_Mat self, void **mat)
{
  PARMS_VALIDMATPTR(self);

  return self->d_vptr->getlmat(self, mat);
}

/** 
 * Get the communication handler.
 * 
 * @param self    A matrix object.                     
 * @param handler The communication handler returned.  
 * 
 * @return 0 on success.
 */
int parms_MatGetCommHandler(parms_Mat self, parms_Comm *handler)
{
  PARMS_VALIDMATPTR(self);

  return self->d_vptr->gethandler(self, handler);
}
