#include "parms_opt_impl.h"
#include "parms_mat_impl.h"
#include "../parms_ilu_vcsr.h"

static int MatView_vcsr(parms_Mat self, parms_Viewer v)
{
  int        i, j, lsize, pid, npro, length, *rja;
  FLOAT      *ra;
  parms_vcsr matrix;
  parms_Map  is;
  FILE       *fp;

  PARMS_VALIDMATPTR(self);
  is      = self->is;
  lsize   = parms_MapGetLocalSize(is);
  pid     = is->pid;
  npro    = is->npro;
  parms_ViewerGetFP(v, &fp);

  fprintf(fp, "There is one processor available\n");
  fprintf(fp, "The format of output local equations is:\n");
  fprintf(fp, "(pid,local_row_index)=>(pid,global_row_index)\n");
  fprintf(fp, "(pid,local_row_index, local_column_index) = value\n"); 
  fprintf(fp, "\n");

  for (i = 0; i < lsize; i++) {
    fprintf(fp, "(%d,%d) => (%d, %d)\n", pid, i, pid, i);
  }
  
  matrix = self->aux_data;
  fprintf(fp, "Local diagonal matrix on processor %d is\n", pid);
  for (i = 0; i < lsize; i++) {
    length = matrix->nnzrow[i];
    rja    = matrix->pj[i];
    ra     = matrix->pa[i];
    for (j = 0; j < length; j++) {
      fprintf(fp, "(%d,%d,%d) = %f  ", pid, i, rja[j], ra[j]);
    }
    fprintf(fp, "\n");
  }
  parms_ViewerStoreFP(v, fp);
  return 0;
}

static int MatFree_vcsr(parms_Mat *self)
{
  return 0;
}

static int MatApply_vcsr(parms_Mat self, parms_Vec x, parms_Vec y)
{
  int          lsize, i, j, index, *pj, length;
  parms_vcsr   matrix;
  parms_Map    is;
  FLOAT        *xvec, *yvec, *pa, s;
  
  PARMS_VALIDMATPTR(self);
  PARMS_VALIDVECPTR(x);
  PARMS_VALIDVECPTR(y);
  
  /* extract diagonal and off-diagonal matrix */
  matrix   = self->aux_data;
  is       = self->is;
  lsize    = parms_MapGetLocalSize(is);

  /* perform local matrix vector product */
  parms_VecGetArray(x, &xvec);
  parms_VecGetArray(y, &yvec);

  for (i = 0; i < lsize; i++) {
    s = 0.0;
    length = matrix->nnzrow[i];
    pj  = matrix->pj[i];
    pa  = matrix->pa[i];
    for (j = 0; j < length; j++) {
      index = pj[j];
      s    += pa[j] * xvec[index];
    }
    yvec[i] = s;
  }

  parms_VecRestoreArray(x, &xvec);
  parms_VecRestoreArray(y, &yvec);

  return 0;
}

static int MatMVPY_vcsr(parms_Mat self, REAL alpha, parms_Vec x,
			REAL beta, parms_Vec y, parms_Vec z) 
{
  int          lsize, i, j, index, *pj, nnz;
  parms_vcsr   matrix;
  parms_Map    is;
  FLOAT        *xvec, *yvec, *zvec, *pa, s;
  
  /* extract diagonal and off-diagonal matrix */
  matrix   = self->aux_data;
  is       = self->is;
  lsize    = is->lsize;

  /* perform local matrix vector product */
  parms_VecGetArray(x, &xvec);
  parms_VecGetArray(y, &yvec);
  parms_VecGetArray(z, &zvec);

  for (i = 0; i < lsize; i++) {
    s = beta * yvec[i];
    nnz = matrix->nnzrow[i];
    pj  = matrix->pj[i];
    pa  = matrix->pa[i];
    for (j = 0; j < nnz; j++) {
      index = pj[j];
      s    += alpha * pa[j] * xvec[index];
    }
    zvec[i] = s;
  }

  parms_VecRestoreArray(x, &xvec);
  parms_VecRestoreArray(y, &yvec);
  parms_VecRestoreArray(z, &zvec);
  
  return 0;
}

static int MatSetup_vcsr(parms_Mat self)
{
  /* free member space in aux_data */
  PARMS_FREE(self->aux_data->space);
  self->issetup = true;
  self->is->ninf_send = 0;
  self->is->schur_start =  self->is->lsize;
  return 0;
}

static int MatSetCommType_vcsr(parms_Mat self, COMMTYPE ctype) 
{
  return 0;
}

static int MatGetDiag_vcsr(parms_Mat self, void **mat)
{
  if (self->d_vptr->ilu == parms_arms_vcsr) {
    parms_vcsr diag, diag_mat;
    int i, nnz;

    diag_mat = self->aux_data;
    PARMS_NEW(diag);
    diag->n = diag_mat->n;
    PARMS_NEWARRAY(diag->nnzrow, diag->n);
    PARMS_NEWARRAY(diag->pj,     diag->n);
    PARMS_NEWARRAY(diag->pa,     diag->n);
    for (i = 0; i < diag->n; i++) {
      nnz = diag->nnzrow[i] = diag_mat->nnzrow[i];
      if (nnz) {
	PARMS_NEWARRAY(diag->pj[i], nnz);
	PARMS_NEWARRAY(diag->pa[i], nnz);
	PARMS_MEMCPY(diag->pj[i], diag_mat->pj[i], nnz);
	PARMS_MEMCPY(diag->pa[i], diag_mat->pa[i], nnz);
      }
    }
    *mat = diag;
    return 0;
  }
  *mat = self->aux_data;
  return 0;
}

static int MatSetILUType_vcsr(parms_Mat self, PCILUTYPE type)
{
  self->d_vptr->ilu = parms_ilu_func_vcsr[type];
  return 0;
}

static struct parms_Mat_ops parms_matops_vtbl = {
  MatApply_vcsr,
  MatSetup_vcsr,
  MatSetCommType_vcsr,
  MatMVPY_vcsr,
  MatGetDiag_vcsr,
  MatGetDiag_vcsr,
  0,
  MatSetILUType_vcsr,
  0,
  0,
  0,
  0
};

int parms_MatCreate_vcsr(parms_Mat self)
{
  self->o_vptr->dtor = (o_dtor)MatFree_vcsr;
  self->o_vptr->view = (o_view)MatView_vcsr;
  PARMS_MEMCPY(self->d_vptr, &parms_matops_vtbl, 1);
  return 0;
}
