#include "parms_comm_impl.h"
#include "parms_pc_impl.h"
#include "parms_opt_impl.h"

typedef struct ras_data {
  parms_Operator op;
  parms_Comm handler;
  parms_Map  is;
  parms_Vec x_ext;
  parms_Vec y_ext;
  FLOAT *rbuf;
  int nloc;
  int n;
  int nodv;
  int nsend;
} *ras_data;

/** Free the memory for struct ras_data.
 *
 *  \param self A preconditioner object.
 *  \return 0 on success.
 */
static int pc_ras_free(parms_PC *self)
{
  ras_data pc_data;
  parms_Operator op;

  PARMS_VALIDPCPTR(*self);
  pc_data = (ras_data)(*self)->data;
  op = pc_data->op;
  parms_OperatorFree(&op);
  parms_MapFree(&pc_data->is);
  parms_VecFree(&pc_data->x_ext);
  parms_VecFree(&pc_data->y_ext);
  PARMS_FREE(pc_data);
  return 0;
}

/** Dump RAS preconditioner.
 *
 *  \param self A preconditioner object.
 *  \param v    A viewer object.
 *  \return 0 on success.
 */
static int pc_ras_view(parms_PC self, parms_Viewer v)
{
  ras_data pc_data;

  PARMS_VALIDPCPTR(self);
  pc_data = (ras_data)self->data;
  parms_OperatorView(pc_data->op, v);
  return 0;
}

/** Set up RAS preconditioner.
 *
 *  \param self  A preconditioner object.
 *  \return 0 on success.
 */
static int pc_ras_setup(parms_PC self)
{
  ras_data pc_data;
  parms_Mat A;
  parms_Comm handler;
  parms_FactParam param;
  parms_Operator op;
  void *lmat, *mat_ext;

  PARMS_VALIDPCPTR(self);

  /* get the matrix */
  A =  self->A;
  /* get communication handler */
  parms_MatGetCommHandler(A, &handler);

  pc_data = (ras_data)self->data;
  pc_data->nloc = A->is->lsize;
  pc_data->handler = handler;
  pc_data->nodv = parms_CommGetNumRecv(handler);
  pc_data->nsend = parms_CommGetNumSend(handler);
  parms_CommGetRecvBuf(handler, &pc_data->rbuf);

  /* get the local matrix */
  parms_MatGetSubMat(A, &lmat);

  /* extend the local matrix by including the equations correspond to 
     the immediate neighbouring variables */
  parms_MatExtend(A, handler, 0, lmat, &pc_data->n, &mat_ext);

  parms_MapCreateFromLocal(&pc_data->is, pc_data->n, 0);
  parms_VecCreate(&pc_data->x_ext, pc_data->is);
  parms_VecCreate(&pc_data->y_ext, pc_data->is);

  /* set parameters used for ILU factorization */
  param = self->param;
  param->start = 0;
  param->n = pc_data->n;
  param->schur_start = param->n;
  param->isalloc = false;
  if (param->ipar[4] != 0) {
    param->ipar[4] = 0;
  }
  if (param->ipar[5] != 0) {
    param->ipar[5] = 0;
  }

  /* ILU factorization */
  parms_MatILU(A, param, mat_ext, &op);

  /* free the memory for the mat_ext */
  if (self->pcilutype != PCARMS) {
    parms_MatFreeSubMat(A, mat_ext);
  }


  pc_data->op = op;
  return 0;
}

/** Apply RAS to vector y.
 *
 *  \f$x = self^{-1}y\f$.
 *
 *  \param self  A RAS preconditioner.
 *  \param y     A right-hand-side vector.
 *  \param x     Solution vector.
 *  \return 0 on success.
 */
static int pc_ras_apply(parms_PC self, parms_Vec y, parms_Vec x)
{
  ras_data pc_data;
  parms_Operator op;
  parms_Mat A;
  parms_Comm handler;
  FLOAT *xvec, *yvec;
  FLOAT *xvec_ext, *yvec_ext;
  FLOAT *rbuf;
  int nodv, nsend, n, nloc;

  PARMS_VALIDPCPTR(self);
  PARMS_VALIDVECPTR(y);
  PARMS_VALIDVECPTR(x);

  pc_data = (ras_data)self->data;
  op = pc_data->op;
  handler = pc_data->handler;
  rbuf = pc_data->rbuf;
  nsend = pc_data->nsend;
  nodv = pc_data->nodv;
  n = pc_data->n;
  nloc = pc_data->nloc;
  A = self->A;

  parms_VecGetArray(y, &yvec);
  parms_VecGetArray(pc_data->y_ext, &yvec_ext);

  /* exchange interface variables */
  if (nsend) {
    parms_CommDataBegin(handler, yvec, 0);
  }
  /* copy local variables to the extended vector */
  PARMS_MEMCPY(yvec_ext, yvec, pc_data->nloc);
  if (nodv) {
    parms_CommDataEnd(handler);
  }
  /* copy received external interface variables to the extended
     vector */
  if (pc_data->n-pc_data->nloc) {
    PARMS_MEMCPY(&yvec_ext[nloc], rbuf, pc_data->n-pc_data->nloc);
  }
  parms_VecRestoreArray(pc_data->y_ext, &yvec_ext);
  /* solve the extended linear system */
  parms_OperatorApply(op, pc_data->y_ext, pc_data->x_ext);
  parms_VecGetArray(pc_data->x_ext, &xvec_ext);
  parms_VecGetArray(x, &xvec);
  /* ignore external variables -RAS */
  PARMS_MEMCPY(xvec, xvec_ext, nloc);
  parms_VecRestoreArray(x, &xvec);
  parms_VecRestoreArray(y, &yvec);
  return 0;
}
/** Get the ratio of the number of nonzero entries of the
 *  preconditioning matrix to that of the original matrix.
 *
 *  \param self   A preconditioner.
 *  \param ratio  A pointer to the ratio.      
 *  \return 0 on success.
 */
static int pc_ras_getratio(parms_PC self, double *ratio)
{
  ras_data pc_data;
  parms_Operator op;
  int nnz_mat, nnz_pc;
  int gnnz_mat, gnnz_pc;

  pc_data = (ras_data)self->data;
  op = pc_data->op;
  parms_OperatorGetNnz(op, &nnz_mat, &nnz_pc);
  MPI_Allreduce(&nnz_mat, &gnnz_mat, 1, MPI_INT, MPI_SUM,
		MPI_COMM_WORLD); 
  MPI_Allreduce(&nnz_pc, &gnnz_pc, 1, MPI_INT, MPI_SUM,
		MPI_COMM_WORLD); 
  *ratio = (double)gnnz_pc/(double)gnnz_mat;  
  return 0;
}


/** Create a RAS preconditioner.
 *
 *  \param self A preconditioner object.
 *  \return 0 on success.
 */
int parms_PCCreate_RAS(parms_PC self)
{
  ras_data data;

  PARMS_VALIDPCPTR(self);
  PARMS_NEW(data);
  self->data = data;
  self->o_vptr->dtor     = (o_dtor)pc_ras_free;
  self->o_vptr->view     = (o_view)pc_ras_view;
  self->d_vptr->apply    = pc_ras_apply;
  self->d_vptr->solve    = pc_ras_apply;
  self->d_vptr->setup    = pc_ras_setup;
  self->d_vptr->getratio = pc_ras_getratio;
  return 0;
}
