/*--------------------------------------------------------------------
  parms_TableCreate   : create a hash table.
  parms_TableFree     : free the memory for the table.
  parms_TableGet      : get the corresponding value for a given key.
  parms_TableGetSize  : get the total number of entries in the table.
  parms_TablePut      : put the pair (key, value) into the table.

  A code fragment for using table functions:

  parms_Table table;
  int key, value, *p;

  // create a hash table using the default hash function
  parms_TableCreate(&table, NULL, tsize);
  // put pair(key, value) into the table
  parms_TablePut(table, key, value);
  // get the corresponding value for a given key.
  p = parms_TableGet(table, key);
  if (p != NULL) {
    value = *p;
  }
  // free the memory for the table object.
  parms_TableFree(&table);
  
  $Id: parms_table.c,v 1.1.1.1 2006/11/24 21:15:31 zli Exp $
  ------------------------------------------------------------------*/

#include "parms_table_impl.h"

static unsigned  hashfunc(int key)
{
  unsigned k;

  k = key;
  return k;
}

/** 
 * Create a hash table.
 *
 * Create a hash table with hf as the hash function.
 * 
 * @param newT  A pointer to the hash table created.              
 * @param hf 	The hash function. If hf is null, the default hash
 *              function is used.     
 * @param size  The number of entries stored in the table.
 * 
 * @return 0 on success.
 */
int parms_TableCreate(parms_Table *newT, HashFcn hf, int tsize)
{
  int i, lwr;
  parms_Table table;
  static int primes[] = {389, 389, 769, 1543, 3079, 6151, 12289,
			 24593, 49157, 98317, 196613, 393241, 786433,
			 1572869, 3145739, 6291469, 12582917,
			 25165843, 50331653, 100663319, 201326611,
			 402653189, 805306457, 1610612741}; 

  /* find the maximum prime less than tsize */
  for (i = 1; lwr = (primes[i]*2)/3, lwr < tsize; i++);

  /* malloc the space for the table */
  table = PARMS_ALLOC(sizeof(*table) +
		      sizeof(table->Slots[0])*primes[i-1]); 
  table->space = primes[i-1];
  table->Slots = (struct Slot **)(table+1);
  for (i = 0; i < table->space; i++) {
    table->Slots[i] = NULL;
  }

  table->hf = hf ? hf : hashfunc;
  table->size = 0;
  *newT = table;
  return 0;
}

/** 
 * Get the corresponding value for a given key.
 *
 * If return NULL, then the entry with key is not in the table,
 * otherwise return a pointer to the value.
 * 
 * @param table A parms_Table object. 
 * @param key   The key value.     
 * 
 * @return A pointer to the value.
 */
void *parms_TableGet(parms_Table table, int key)
{
  int index;
  struct Slot *p;
  
  index = table->hf(key) % table->space;
  for (p = table->Slots[index]; p; p = p->link) {
    if (p->key == key) {
      break;
    }
  }
  return p ? &p->value : NULL;
}

/** 
 * Put the pair (key, value) into the table.
 * 
 * @param table A parms_Table object.        
 * @param key 	The key of the pair.         
 * @param val 	The value of the pair.       
 * 
 * @return 0 on success.
 */
int parms_TablePut(parms_Table self, int key, int value)
{
  int index;
  struct Slot *p;

  index = self->hf(key) % self->space;
  for (p = self->Slots[index]; p; p = p->link) {
    if (p->key == key) {
      break;
    }
  }
  if (p == NULL) {		/* new entry */
    PARMS_NEW(p);
    p->key = key;
    p->value = value;
    p->link = self->Slots[index];
    self->Slots[index] = p;
    self->size++;
  }
  else {
    p->key = key;
    p->value = value;
  }
  return 0;
}

/** 
 * Free the memory for the table object pointed by self.
 * 
 * @param table A pointer to the parms_Table object.  
 * 
 * @return 0 on success.
 */
int parms_TableFree(parms_Table *self)
{
  struct Slot *p1, *p2;
  int i;

  for (i = 0; i < (*self)->space; i++) {
    for (p1 = (*self)->Slots[i]; p1; p1 = p2) {
      p2 = p1->link;
      PARMS_FREE(p1);
    }
  }
  PARMS_FREE(*self);
  return 0;
}

/** 
 * Get the number of entries stored in the table.
 * 
 * @param table A hash table object.
 * 
 * @return The number of entries stored in the table.
 */
int parms_TableGetSize(parms_Table self)
{
  return self->size;
}
